(* Mathematica Package *)

BeginPackage["AtomicDensityMatrix`WignerRotate`", 
	{ "AtomicDensityMatrix`",
	"AtomicDensityMatrix`Common`",
	"AtomicDensityMatrix`TensorAlgebra`",
	"AtomicDensityMatrix`AtomicSystem`",
	"AtomicDensityMatrix`WignerEckart`"}]

ADMWignerD::usage=""

WignerD::usage="WignerD[{j,m,m'},{alpha,beta,gamma}] returns the Wigner D-function (matrix element). WignerD[j,{alpha,beta,gamma}] returns the rotation matrix for a system with angular momentum j. WignerD[s1,{alpha,beta,gamma},s2] returns the rotation matrix element between the atomic states s1 and s2. WignerD[syst,{alpha,beta,gamma}] returns the rotation matrix for the atomic system syst."

WignerDLargeJ::usage=""

Wignerd::usage=""

WignerRotate::usage="WignerRotate[tensor, {alpha,beta,gamma}, opts] rotates tensor by the specified Euler angles. The supplied tensor can be a matrix operator, an irreducible tensor, or a tensor decomposition."

RotationSense::usage="RotationSense is an option for WignerRotate that specifies the rotation sense. Possible values are Active and Passive."

Active::usage="Active is a possible value for the RotationSense option for WignerRotate, specifying an active rotation."

Passive::usage="Passive is a possible value for the RotationSense option for WignerRotate, specifying a passive rotation."

StandardOperator::usage="StandardOperator is an option for WignerRotate that specifies whether objects are considered to rotate like standard operators, as opposed to density matrices."

Begin["`Private`"] (* Begin Private Context *) 

(*Wigner D-function--main definition*)

(*Matrix form*)

DownValues[WignerD] = Join[{HoldPattern[WignerD[x_, {al_,be_,ga_}, z___]] :> ADMWignerD[x, {al,be,ga}, z]},DownValues[WignerD]]

ADMWignerD[s1_AtomicState,{al_,be_,ga_},s2_AtomicState]:=
	KroneckerDelta[StateLabel[s1],StateLabel[s2]]Switch[AtomicSystemType[s1],
	"Toy",1,
	"Zeeman"|"HyperfineZeeman",KroneckerDelta[F[s1],F[s2]]ADMWignerD[{F[s1],M[s1],M[s2]},{al,be,ga}]
]

ADMWignerD[system:{_AtomicState..},{al_,be_,ga_},opts___Rule]:=DirectSelfProduct[ADMWignerD[#1,{al,be,ga},#2]&,system]

ADMWignerD[J_?NumericQ,{al_,be_,ga_}]:=ADMWignerD[ZeemanSystem[J],{al,be,ga}]

(*Special case \[Beta]=n \[Pi], n even  [Var88, Sec. 4.16, Eq. (3)].*)

ADMWignerD[{J_,m_,mp_},{al_,(\[Pi] n_./;EvenQ[n])|n_/;n==0,ga_}]:=KroneckerDelta[m,mp] (-1)^(n J) E^(-I m (al+ga))

(*Special case \[Beta]=n \[Pi], n odd  [Var88, Sec. 4.16, Eq. (4)].*)

ADMWignerD[{J_,m_,mp_},{al_,\[Pi]  n_./;OddQ[n],ga_}]:=KroneckerDelta[-m,mp] (-1)^(n J+m) E^(-I m (al-ga))

(*Definition in terms of d [Var88, Sec. 4.3, Eq. (1)]*)

ADMWignerD[{J_?NumericQ,m_?NumericQ,mp_?NumericQ},{al_,be_,ga_}]:=E^(-I m al) Wignerd[{J,m,mp},be] E^(-I mp ga)

(*Definition of d, from spin-1/2construction [Cha98, Eq. (6.2.16)]*)

Wignerd[{j_,m1_,m_},be_]:=Wignerd[{j,m1,m},be]=Sqrt[((j+m1)!(j-m1)!)/((j+m)!(j-m)!)] Cos[be/2]^(-m-m1) Sin[be/2]^(2j+m+m1) (-1)^(j+m1) Sum[((-1)^\[Nu])Binomial[j+m,\[Nu]]Binomial[j-m,j+m1-\[Nu]] Cos[be/2]^(2\[Nu]) Sin[be/2]^(-2\[Nu]),{\[Nu],Max[0,m+m1],Min[j+m,j+m1]}]

(*Wigner-D--additional definitions*)

(*Special case M=0. [Var88, Sec. 4.17 (a), Eq. (1)]*)

(* ::Input:: *)
(*WignerD1[{J_,0,MP_},{al_,be_,ga_}]:=Sqrt[(4\[Pi])/(2 J+1)]SphericalHarmonicY[J,-MP,be,ga]*)

(* ::Input:: *)
(*WignerD1[{J_,M_,0},{al_,be_,ga_}]:=(-1)^M Sqrt[(4\[Pi])/(2 J+1)]SphericalHarmonicY[J,-M,be,al]*)

(*Special case M=+/-J [Var88, Sec. 4.17 (e), Eq. (8)]*)

(* ::Input:: *)
(*WignerD1[{J_,m_,-J_},{al_,be_,ga_}]:=(-1)^(J+m) Sqrt[(2J)!/((J+m)!(J-m)!)] Cos[be/2]^(J-m) Sin[be/2]^(J+m) E^(-I m al+I J ga)*)

(* ::Input:: *)
(*WignerD1[{J_,m_,J_},{al_,be_,ga_}]:=Sqrt[(2J)!/((J+m)!(J-m)!)] Cos[be/2]^(J+m) Sin[be/2]^(J-m) E^(-I m al-I J ga)*)

(*Asymptotic for large J [Var88, Sec. 4.18.1, Eq. (1)] *)

(* ::Input:: *)
(*WignerDLargeJ[{J_,m_,mp_},{al_,be_,ga_}]=E^(-I m al-I mp ga) \[Xi][m,mp] Sqrt[(s!(s+\[Mu] +\[Nu])!)/((s+\[Mu])!(s+\[Nu])!)] Sqrt[2/(\[Pi] s)] Cos[(s+(\[Mu]+\[Nu]+1)/2)be-\[Pi]/4 (2\[Mu]+1)]/Sqrt[Sin[be]]//.{\[Mu]->m-mp,\[Nu]->m+mp,s->J-1/2 (\[Mu]+\[Nu]),\[Xi][m_,mp_]->(-1)^(mp-m)}*)

(*Definition of d [Var88, Sec. 4.3.1, Eq. (4)]*)

(*Wignerd[{J_,M_,MP_},\[CapitalTheta]_]:=Wignerd[{J,M,MP},\[CapitalTheta]]=((J+M)! (J-M)! (J+MP)! (J-MP)!)^(1/2) \!\(
\*UnderoverscriptBox[\(\[Sum]\), \(k = Max[0, M - MP]\), \(Min[J - MP, J + M]\)]\(SuperscriptBox[\((\(-1\))\), k] 
\*FractionBox[\(SuperscriptBox[Cos[\[CapitalTheta]/2], 2\ J - 2  k + M - MP] 
\*SuperscriptBox[\(Sin[\[CapitalTheta]/2]\), \(2  k - M + MP\)]\), \(\(k!\) \(\((J + M - k)\)!\)\ \(\((J - MP - k)\)!\)\ \(\((MP - M + k)\)!\)\)]\)\)*)

(*Rotations*)

(*RotateCoordinates[tensor, {al,be,ga}] returns tensor in a new coordinate system obtained by rotating the original according to the Euler angles al, be, ga. The supplied tensor can be a reducible tensor, an irreducible tensor, or a list of irreducible tensors describing the decomposition of a reducible tensor.*)

(*All components of an irreducible tensor in the new basis. [Var88, Sec. 3.1.3, Eq. (11); Sec. 4.1, Eq. (2)].*)
(*	|\[Psi](X')\[RightAngleBracket]=D|\[Psi](X)\[RightAngleBracket]*)
(*	(((|\[Psi](X')\[RightAngleBracket]))^\[Dagger]=(D|\[Psi](X)\[RightAngleBracket]))^\[Dagger]*)
(*	\[LeftAngleBracket]\[Psi](X')|=\[LeftAngleBracket]\[Psi](X)|D^\[Dagger]*)

(*The following definition applies to both irreducible tensors and irreducible tensor operators. [Var88, Sec. 4.1 Eq. (2)], [Sak85, Eq. (3.10.22b)].*)

Options[WignerRotate]={RotationSense->Active,StandardOperator->False};

WignerRotate[params__,{al_,be_,ga_},opts___Rule]/;(RotationSense/.{opts}/.Options[WignerRotate])===Passive:=
	WignerRotate[params,{-ga,-be,-al},RotationSense->Active]

WignerRotate[T_?TensorDecompositionQ,{al_,be_,ga_},opts___Rule]:=
	WignerRotate[#,{al,be,ga},opts]&/@T

WignerRotate[T_,{al_,be_,ga_},opts___Rule]:=
	WignerRotate[ZeemanSystem[Rank[T]],T,{al,be,ga},opts]

WignerRotate[syst:{_AtomicState..},T_?OneByOneQ,{al_,be_,ga_},opts___Rule]:=T

(*Rotate an irreducible tensor. If we want to rotate the coefficients of an expansion in an irreducible tensor, we need to apply a rotation in the opposite sense.*)

WignerRotate[syst:{_AtomicState..},T_?CovariantQ,{al_,be_,ga_},OptionsPattern[]]:=
	If[OptionValue[StandardOperator]&&!SphericalTensorQ[T,OperatorQ],
		T\[CenterDot]WignerD[syst,{al,be,ga}],		
		T\[CenterDot]ECT@WignerD[syst,{al,be,ga}]
	]

WignerRotate[syst:{_AtomicState..},T_?ContravariantQ,{al_,be_,ga_},OptionsPattern[]]:=
	If[OptionValue[StandardOperator]&&!SphericalTensorQ[T,OperatorQ],
		ECT@WignerD[syst,{al,be,ga}]\[CenterDot]T,
		WignerD[syst,{al,be,ga}]\[CenterDot]T
	]

(*Rotate a density matrix in the m m' basis. [Var88, Sec. 1.4.5, Eq. (28) (note typo); Sec. 4.1., Eq. (4)]
Note that a general operator (observable) rotates as the hermitian conjugate of this expression.*)

WignerRotate[syst:{_AtomicState..},T_?OperatorQ,{al_,be_,ga_},OptionsPattern[]]:=
	If[OptionValue[StandardOperator],
		ECT@WignerD[syst,{al,be,ga}]\[CenterDot]T\[CenterDot]WignerD[syst,{al,be,ga}],
		WignerD[syst,{al,be,ga}]\[CenterDot]T\[CenterDot]ECT@WignerD[syst,{al,be,ga}]
	]


End[] (* End Private Context *)

EndPackage[]